from typing import Any
import os
from eth_typing import Address
from eth_utils.address import to_checksum_address
from web3 import Web3
from web3.contract.contract import Contract
from web3.middleware.geth_poa import geth_poa_middleware
from web3.types import TxParams

from ...contract.abi.imageRegistryAbi import contract
from ...enums import ECAddress, ECRunner, ECNetworkRPCDictionary


class ImageRegistryContract:
    def __init__(
        self,
        network_address: Address = ECAddress.BLOXBERG.TESTNET_ADDRESS,  # type: ignore
        runner_type: str = ECRunner.BLOXBERG.NODENITHY_RUNNER,
        signer: Any = None,
    ):
        self.signer = signer
        self.provider = self.newProvider(ECNetworkRPCDictionary[network_address])
        if network_address == ECAddress.BLOXBERG.TESTNET_ADDRESS:
            if runner_type == ECRunner.BLOXBERG.NODENITHY_RUNNER_TESTNET:
                self.contract = self.provider.eth.contract(
                    address=to_checksum_address(
                        ECAddress.BLOXBERG.IMAGE_REGISTRY.NODENITHY.TESTNET_ADDRESS
                    ),
                    abi=contract["abi"],
                )
                # self.provider.eth.contract()
            elif runner_type == ECRunner.BLOXBERG.PYNITHY_RUNNER_TESTNET:
                self.contract = self.provider.eth.contract(
                    address=to_checksum_address(
                        ECAddress.BLOXBERG.IMAGE_REGISTRY.PYNITHY.TESTNET_ADDRESS
                    ),
                    abi=contract["abi"],
                )
        elif network_address == ECAddress.BLOXBERG.MAINNET_ADDRESS:
            if runner_type == ECRunner.BLOXBERG.NODENITHY_RUNNER:
                self.contract = self.provider.eth.contract(
                    address=to_checksum_address(
                        ECAddress.BLOXBERG.IMAGE_REGISTRY.NODENITHY.MAINNET_ADDRESS
                    ),
                    abi=contract["abi"],
                )
            elif runner_type == ECRunner.BLOXBERG.PYNITHY_RUNNER:
                self.contract = self.provider.eth.contract(
                    address=to_checksum_address(
                        ECAddress.BLOXBERG.IMAGE_REGISTRY.PYNITHY.MAINNET_ADDRESS
                    ),
                    abi=contract["abi"],
                )
        elif network_address == ECAddress.POLYGON.MAINNET_ADDRESS:
            if runner_type == ECRunner.POLYGON.NODENITHY_RUNNER:
                self.contract = self.provider.eth.contract(
                    address=to_checksum_address(
                        ECAddress.POLYGON.IMAGE_REGISTRY.NODENITHY.MAINNET_ADDRESS
                    ),
                    abi=contract["abi"],
                )
            elif runner_type == ECRunner.POLYGON.PYNITHY_RUNNER:
                self.contract = self.provider.eth.contract(
                    address=to_checksum_address(
                        ECAddress.POLYGON.IMAGE_REGISTRY.PYNITHY.MAINNET_ADDRESS
                    ),
                    abi=contract["abi"],
                )
        elif network_address == ECAddress.POLYGON.TESTNET_ADDRESS:
            if runner_type == ECRunner.POLYGON.NODENITHY_RUNNER_TESTNET:
                self.contract = self.provider.eth.contract(
                    address=to_checksum_address(
                        ECAddress.POLYGON.IMAGE_REGISTRY.NODENITHY.TESTNET_ADDRESS
                    ),
                    abi=contract["abi"],
                )
            elif runner_type == ECRunner.POLYGON.PYNITHY_RUNNER_TESTNET:
                self.contract = self.provider.eth.contract(
                    address=to_checksum_address(
                        ECAddress.POLYGON.IMAGE_REGISTRY.PYNITHY.TESTNET_ADDRESS
                    ),
                    abi=contract["abi"],
                )

    def newProvider(self, url: str) -> Web3:
        _w3 = Web3(Web3.HTTPProvider(url))
        _w3.enable_unstable_package_management_api()
        _w3.middleware_onion.inject(geth_poa_middleware, layer=0)
        return _w3

    def get_signer(self) -> Any:
        return self.signer

    def get_contract(self) -> Contract:
        return self.contract

    def get_provider(self) -> Web3:
        return self.provider

    def get_enclave_details_v3(
        self, image_name: str, version: str, trustedZoneImage: str = ""
    ) -> Any:
        try:
            if not trustedZoneImage or image_name == trustedZoneImage:
                return self.contract.functions.getLatestTrustedZoneImageCertPublicKey(
                    image_name, version
                ).call()
            else:
                trustedZonePublicKey = (
                    self.contract.functions.getLatestTrustedZoneImageCertPublicKey(
                        trustedZoneImage, "v3"
                    ).call()[1]
                )
                imageDetails = self.contract.functions.getLatestImageVersionPublicKey(
                    image_name, os.environ.get("VERSION")
                ).call()
                return [imageDetails[0], trustedZonePublicKey, imageDetails[2]]
        except Exception as e:
            print(e)
            return None



from enum import Enum
from enum import IntEnum

class ECStatus(Enum):
    ERROR: str = "error"
    SUCCESS: str = "success"
    DEFAULT: str = ""


class ECEvent(Enum):
    TASK_NEW: str = "ecTaskNew"
    TASK_CREATED: str = "ecTaskCreated"
    TASK_ORDER_PLACED: str = "ecTaskOrderPlaced"
    TASK_NOT_PROCESSED: str = "ecTaskNotProcessed"
    TASK_PROGRESS: str = "ecTaskProgress"
    TASK_COMPLETED: str = "ecTaskCompleted"


ECOrderTaskStatus = {
    0: "SUCCESS",
    1: "SYSTEM_ERROR",
    2: "KEY_ERROR",
    3: "SYNTAX_WARNING",
    4: "BASE_EXCEPTION",
    5: "PAYLOAD_NOT_DEFINED",
    6: "PAYLOAD_CHECKSUM_ERROR",
    7: "INPUT_CHECKSUM_ERROR",
}


class ECOrderTaskStatusCode(Enum):
    SUCCESS = "SUCCESS"
    SYSTEM_ERROR = "SYSTEM_ERROR"
    KEY_ERROR = "KEY_ERROR"
    SYNTAX_WARNING = "SYNTAX_WARNING"
    BASE_EXCEPTION = "BASE_EXCEPTION"
    PAYLOAD_NOT_DEFINED = "PAYLOAD_NOT_DEFINED"
    PAYLOAD_CHECKSUM_ERROR = "PAYLOAD_CHECKSUM_ERROR"
    INPUT_CHECKSUM_ERROR = "INPUT_CHECKSUM_ERROR"


ECNetworkByChainId = {
    "BLOXBERG": {"TESTNET": 8995, "MAINNET": 8995},
    "POLYGON": {"TESTNET": 80001, "MAINNET": 137},
}

ECNetworkByChainIdDictionary = {8995: "bloxberg", 80001: "maticmum", 137: "matic"}


class ECNetwork:
    class BLOXBERG:
        TESTNET = "Bloxberg TESTNET"
        MAINNET = "Bloxberg MAINNET"

    class POLYGON:
        TESTNET = "Polygon TESTNET"
        MAINNET = "Polygon MAINNET"


class ECRunner:
    class BLOXBERG:
        PYNITHY_RUNNER_TESTNET = "etny-pynithy-testnet"
        NODENITHY_RUNNER_TESTNET = "etny-nodenithy-testnet"
        PYNITHY_RUNNER = "etny-pynithy"
        NODENITHY_RUNNER = "etny-nodenithy"

    class POLYGON:
        PYNITHY_RUNNER_TESTNET = "ecld-pynithy-mumbai"
        NODENITHY_RUNNER_TESTNET = "ecld-nodenithy"
        PYNITHY_RUNNER = "ecld-pynithy"
        NODENITHY_RUNNER = "ecld-nodenithy"


class ECAddress:
    class BLOXBERG:
        TESTNET_ADDRESS = "0x02882F03097fE8cD31afbdFbB5D72a498B41112c"
        MAINNET_ADDRESS = "0x549A6E06BB2084100148D50F51CF77a3436C3Ae7"

        class IMAGE_REGISTRY:
            class PYNITHY:
                TESTNET_ADDRESS = "0x15D73a742529C3fb11f3FA32EF7f0CC3870ACA31"
                MAINNET_ADDRESS = "0x15D73a742529C3fb11f3FA32EF7f0CC3870ACA31"

            class NODENITHY:
                TESTNET_ADDRESS = "0x15D73a742529C3fb11f3FA32EF7f0CC3870ACA31"
                MAINNET_ADDRESS = "0x15D73a742529C3fb11f3FA32EF7f0CC3870ACA31"

    class POLYGON:
        TESTNET_ADDRESS = "0xfb450e40f590F1B5A119a4B82E6F3579D6742a00"
        TESTNET_PROTOCOL_ADDRESS = "0x4274b1188ABCfa0d864aFdeD86bF9545B020dCDf"
        MAINNET_ADDRESS = "0xc6920888988cAcEeA7ACCA0c96f2D65b05eE22Ba"
        MAINNET_PROTOCOL_ADDRESS = "0x439945BE73fD86fcC172179021991E96Beff3Cc4"

        class IMAGE_REGISTRY:
            class PYNITHY:
                TESTNET_ADDRESS = "0xF7F4eEb3d9a64387F4AcEb6d521b948E6E2fB049"
                MAINNET_ADDRESS = "0x689f3806874d3c8A973f419a4eB24e6fBA7E830F"

            class NODENITHY:
                TESTNET_ADDRESS = "0xF7F4eEb3d9a64387F4AcEb6d521b948E6E2fB049"
                MAINNET_ADDRESS = "0x689f3806874d3c8A973f419a4eB24e6fBA7E830F"


class ECNetworkName(Enum):
    BLOXBERG = "bloxberg"
    MUMBAI = "maticmum"
    POLYGON = "matic"
    OTHER = "other"


ECNetworkNameDictionary = {
    ECAddress.BLOXBERG.MAINNET_ADDRESS: "bloxberg",
    ECAddress.BLOXBERG.TESTNET_ADDRESS: "bloxberg",
    ECAddress.POLYGON.TESTNET_ADDRESS: "maticmum",
    ECAddress.POLYGON.MAINNET_ADDRESS: "matic",
}

ECNetworkName1Dictionary = {
    ECAddress.BLOXBERG.MAINNET_ADDRESS: "BLOXBERG",
    ECAddress.BLOXBERG.TESTNET_ADDRESS: "BLOXBERG",
    ECAddress.POLYGON.TESTNET_ADDRESS: "MUMBAI",
    ECAddress.POLYGON.MAINNET_ADDRESS: "POLYGON",
}

ECNetworkRPCDictionary = {
    ECAddress.BLOXBERG.MAINNET_ADDRESS: "https://core.bloxberg.org",
    ECAddress.BLOXBERG.TESTNET_ADDRESS: "https://core.bloxberg.org",
    ECAddress.POLYGON.TESTNET_ADDRESS: "https://rpc-mumbai.matic.today",
    ECAddress.POLYGON.MAINNET_ADDRESS: "https://rpc-mainnet.maticvigil.com",
}

ECNetworkEnvToEnum = {
    "bloxberg_mainnet": ECAddress.BLOXBERG.MAINNET_ADDRESS,
    "bloxberg_testnet": ECAddress.BLOXBERG.TESTNET_ADDRESS,
    "polygon_testnet": ECAddress.POLYGON.TESTNET_ADDRESS,
    "polygon_mainnet": ECAddress.POLYGON.MAINNET_ADDRESS,
}

ZERO_CHECKSUM = "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855"


class ECError(Enum):
    PARSE_ERROR = "EtnyParseError"
    IPFS_DOWNLOAD_ERROR = "ECIPFSDownloadError"

class ECLog(IntEnum):
    ERROR = 1
    WARNING = 2
    INFO = 3
    DEBUG = 4


import requests  # type: ignore


class IPFSClient:
    def __init__(
        self, api_url: str = "http://ipfs.ethernity.cloud:5001/api/v0", token: str = ""
    ) -> None:
        self.api_url = api_url
        self.headers = {}
        if token:
            self.headers = {"authorization": token}

    def upload_file(self, file_path: str) -> None:
        add_url = f"{self.api_url}/add"

        with open(file_path, "rb") as file:
            files = {"file": file}
            response = requests.post(add_url, files=files, headers=self.headers)

        if response.status_code == 200:
            try:
                response_data = response.json()
                ipfs_hash = response_data["Hash"]
                print(f"Successfully uploaded to IPFS. Hash: {ipfs_hash}")
                return ipfs_hash
            except Exception as e:
                print(f"Failed to upload to IPFS. Error: {e}")
                return None
        else:
            print(f"Failed to upload to IPFS. Status code: {response.status_code}")
            print(response.text)
            return None

    def upload_to_ipfs(self, data: str) -> None:
        add_url = f"{self.api_url}/add"
        files = {"file": data}
        response = requests.post(add_url, files=files, headers=self.headers)

        if response.status_code == 200:
            try:
                response_data = response.json()
                ipfs_hash = response_data["Hash"]
                #print(f"Successfully uploaded to IPFS. Hash: {ipfs_hash}")
                return ipfs_hash
            except Exception as e:
                #print(f"Failed to upload to IPFS. Error: {e}")
                return None
        else:
            print(f"Failed to upload to IPFS. Status code: {response.status_code}")
            print(response.text)
            return None

    def download_file(
        self, ipfs_hash: str, download_path: str, attempt: int = 0
    ) -> None:
        gateway_url = f"https://ipfs.io/ipfs/{ipfs_hash}"
        response = requests.get(url=gateway_url, timeout=60, headers=self.headers)

        if response.status_code == 200:
            with open(download_path, "wb") as file:
                file.write(response.content)
            print(f"File downloaded successfully to {download_path}")
        else:
            print(
                f"Failed to download from IPFS. Attempt {attempt}. Status code: {response.status_code}. Response text: {response.text}.\n{'Trying again...' if attempt < 6 else ''}"
            )
            if attempt < 6:
                self.download_file(ipfs_hash, download_path, attempt + 1)

    def get_file_content(self, ipfs_hash: str, attempt: int = 0) -> None:
        url = self.api_url
        gateway_url = f"{url}/cat?arg={ipfs_hash}"
        response = requests.post(url=gateway_url, timeout=60, headers=self.headers)

        if response.status_code == 200:
            # TODO: use a get encoding function to determine the encoding
            return response.content.decode("utf-8")
        else:
            print(
                f"Failed to get content from IPFS. Attempt {attempt}. Status code: {response.status_code}. Response text: {response.text}.\n{'Trying again...' if attempt < 6 else ''}"
            )
            if attempt < 6:
                self.get_file_content(ipfs_hash, attempt + 1)

        return None


# Example usage
if __name__ == "__main__":
    file_path = "./input.txt"  # Replace with your file path
    download_path = "./downloaded_file.txt"  # Replace with your desired download path

    ipfs_client = IPFSClient()

    # Upload file to IPFS
    ipfs_hash = ipfs_client.upload_file(file_path)

    if ipfs_hash:
        # Download file from IPFS
        ipfs_client.download_file(ipfs_hash, download_path)



from datetime import datetime
import random
import time
from typing import Any
from eth_account import Account
from web3 import Web3
from eth_utils.address import to_checksum_address

from eth_utils.hexadecimal import decode_hex, encode_hex

from .crypto import sha256

# from eth_abi import decode_abi
from eth_utils import function_signature_to_4byte_selector


from eth_account import Account
from eth_account._utils.legacy_transactions import (
    serializable_unsigned_transaction_from_dict,
)
from eth_utils import to_hex, decode_hex


def delay(seconds: int) -> None:
    return time.sleep(seconds)


def get_retry_delay(retry_count: int, base_delay: int = 1) -> int:
    return base_delay * 2**retry_count


def format_date(dt: datetime = datetime.now()) -> str:
    return dt.strftime("%d/%m/%Y %H:%M:%S")


def generate_random_hex_of_size(size: int) -> str:
    return "".join(random.choice("0123456789abcdef") for _ in range(size))


def is_null_or_empty(value: str) -> bool:
    return value is None or value == ""


def generate_wallet(client_challenge: str, enclave_challenge: str) -> Any:
    try:
        encoded = client_challenge + enclave_challenge
        hash = sha256(sha256(encoded, True), True)
        wallet = Account.from_key(hash)
        return wallet.address
    except Exception as e:
        print(e)
        return False


# def generate_wallet(challenge_hash, enclave_challenge):
#     private_key_bytes = hashlib.sha256((challenge_hash + enclave_challenge).encode()).digest()
#     private_key = keys.PrivateKey(private_key_bytes)
#     return private_key.public_key.to_checksum_address()


def is_address(address: str) -> bool:
    try:
        checksum_address = to_checksum_address(address)
    except Exception as e:
        return False
    return True


# def parse_transaction_bytes_ut(contract_abi: Any, bytes_input: bytes) -> Any:
#     contract = w3.eth.contract(abi=contract_abi)
#     decoded_data = contract.decode_function_input(bytes_input)
#     return decoded_data


# def parse_transaction_bytes_ut(contract_abi, bytes_input):
#     # Connect to a Web3 provider (e.g., Infura, local node)
#     w3 = Web3(
#         Web3.HTTPProvider("https://core.bloxberg.org")
#     )  # Replace with your provider

#     # Parse the transaction
#     parsed_transaction = w3.eth.account.decode_transaction(bytes_input)
#     parsed_transaction = w3.eth.account.recover_transaction(bytes_input)
#     # Create a contract interface
#     contract_interface = w3.eth.contract(abi=contract_abi)

#     # Find the function signature
#     function_selector = parsed_transaction.input[:10]
#     function_abi = next(
#         (
#             item
#             for item in contract_abi
#             if function_signature_to_4byte_selector(item["name"]) == function_selector
#         ),
#         None,
#     )

#     if function_abi is None:
#         raise ValueError("Function ABI not found for the given selector")

#     # Decode the transaction data
#     decoded_data = contract_interface.decode_function_input(parsed_transaction.input)

#     return {
#         "from": parsed_transaction["from"],
#         "result": decoded_data[1][1],  # Assuming the result is the second argument
#     }


def parse_transaction_bytes_ut(contract_abi, bytes_input):
    import rlp
    from rlp.sedes import big_endian_int, Binary, binary
    from eth_utils import keccak, to_checksum_address, decode_hex
    from eth_keys import keys
    from web3 import Web3

    # Define the signed transaction class
    class SignedTransaction(rlp.Serializable):
        fields = [
            ("nonce", big_endian_int),
            ("gasPrice", big_endian_int),
            ("gas", big_endian_int),
            ("to", Binary.fixed_length(20, allow_empty=True)),
            ("value", big_endian_int),
            ("data", binary),
            ("v", big_endian_int),
            ("r", big_endian_int),
            ("s", big_endian_int),
        ]

    # Define the unsigned transaction class
    class UnsignedTransaction(rlp.Serializable):
        fields = [
            ("nonce", big_endian_int),
            ("gasPrice", big_endian_int),
            ("gas", big_endian_int),
            ("to", Binary.fixed_length(20, allow_empty=True)),
            ("value", big_endian_int),
            ("data", binary),
        ]

    # Convert hex string to bytes if necessary
    if isinstance(bytes_input, str):
        bytes_input = bytes_input.strip()
        if bytes_input.startswith("0x"):
            bytes_input = decode_hex(bytes_input)
        else:
            bytes_input = bytes.fromhex(bytes_input)

    # Decode the transaction using RLP
    try:
        tx = rlp.decode(bytes_input, SignedTransaction)
    except Exception as e:
        print(f"Error decoding transaction: {e}")
        return None

    # Create an unsigned transaction instance
    unsigned_tx = UnsignedTransaction(
        nonce=tx.nonce,
        gasPrice=tx.gasPrice,
        gas=tx.gas,
        to=tx.to,
        value=tx.value,
        data=tx.data,
    )
    # Create a Web3 instance
    w3 = Web3(Web3.HTTPProvider("https://core.bloxberg.org"))
    # Compute the transaction hash (the message hash used for signing)
    tx_hash = keccak(rlp.encode(unsigned_tx))

    # Recover the sender's public key and address
    v = tx.v
    if v >= 35:
        # EIP-155
        chain_id = (v - 35) // 2
        v_standard = v - (chain_id * 2 + 35) + 27
    else:
        chain_id = None
        v_standard = v

    try:
        # Build the signature object
        # signature = keys.Signature(vrs=(v_standard, tx.r, tx.s))

        # Recover the public key
        # public_key = signature.recover_public_key_from_msg_hash(tx_hash)

        sender_address = w3.eth.account.recover_transaction(bytes_input)
    except Exception as e:
        print(f"Error recovering sender address: {e}")
        return None

    # Decode the function input data
    try:
        contract = w3.eth.contract(abi=contract_abi)
        decoded_function = contract.decode_function_input(tx.data)
        function_name = decoded_function[0].fn_name
        params = decoded_function[1]
    except Exception as e:
        print(f"Error decoding function input: {e}")
        return None

    # Prepare the result
    result = {
        "from": sender_address,
        "to": to_checksum_address(tx.to) if tx.to else None,
        "nonce": tx.nonce,
        "gasPrice": tx.gasPrice,
        "gas": tx.gas,
        "value": tx.value,
        "function_name": function_name,
        "params": params,
        "transaction_hash": "0x" + keccak(bytes_input).hex(),
        "result": params["_result"] if "_result" in params else None,
    }

    return result



from decimal import Decimal
import os
from typing import Any

from eth_account.messages import encode_defunct
from eth_typing import Address, HexStr
from web3 import Web3
from web3.contract.contract import Contract
from web3.middleware.geth_poa import geth_poa_middleware
from web3.types import TxParams

from ..abi.polygonAbi import contract as polygonAbi
from ...enums import ECNetworkByChainIdDictionary, ECNetworkRPCDictionary


class PolygonProtocolContract:
    def __init__(
        self, network_address: Address, signer: Any, is_main_net: bool = False
    ) -> None:
        self.is_main_net = is_main_net
        self.network_address = network_address
        _rpc_url = ECNetworkRPCDictionary[network_address]
        self.provider = Web3(Web3.HTTPProvider(_rpc_url))
        self.provider.enable_unstable_package_management_api()
        self.provider.middleware_onion.inject(geth_poa_middleware, layer=0)

        self.signer = signer
        self.ethernity_contract = self.provider.eth.contract(
            address=network_address, abi=polygonAbi["abi"]
        )

    def contract_address(self) -> Address:
        return self.network_address

    @property
    def __transaction_object(self) -> TxParams:
        nonce = self.provider.eth.get_transaction_count(self.signer.address)
        return {
            "gas": 40500500010 if self.is_main_net else 1300000010,  # type: ignore
            "chainId": 137 if self.is_main_net else 80002,
            "nonce": nonce,
            "gasPrice": self.provider.to_wei("1", "mwei"),  # type: ignore
        }

    def get_signer(self) -> Any:
        return self.signer

    def get_contract(self) -> Contract:
        return self.ethernity_contract

    def get_provider(self) -> Web3:
        return self.provider

    def get_current_wallet(self) -> Any:
        return self.signer

    def get_balance(self) -> int:
        try:
            address = self.signer.address
            balance = self.ethernity_contract.functions.balanceOf(address).call()
            return Web3.from_wei(balance, "ether")
        except Exception as e:
            print(e)
            return 0

    def check_and_set_allowance(
        self, protocol_address: Address, amount: str, task_price: str
    ) -> bool:
        allowance_amount = Web3.to_wei(amount, "ether")
        task_price_amount = Web3.to_wei(task_price, "ether")
        current_wallet_address = self.signer.address
        allowance = self.ethernity_contract.functions.allowance(
            current_wallet_address, protocol_address
        ).call()
        if allowance < task_price_amount:
            tx = self.ethernity_contract.functions.approve(
                protocol_address, allowance_amount
            ).build_transaction(self.__transaction_object)
            signed_tx = self.provider.eth.account.sign_transaction(
                tx, private_key=self.signer._private_key
            )
            self.provider.eth.send_raw_transaction(signed_tx.rawTransaction)
            recept = self.provider.to_hex(
                self.provider.keccak(signed_tx.rawTransaction)
            )
            try:
                self.provider.eth.wait_for_transaction_receipt(recept)
                allowance = self.ethernity_contract.functions.allowance(
                    current_wallet_address, protocol_address
                ).call()
            except Exception as e:
                print(e)
                return False
        return True

    def get_eip1559_gas_options(self) -> dict[str, int]:
        max_fee_per_gas = int(os.getenv("MAX_FEE_PER_GAS", 0)) * 10**9
        max_priority_fee_per_gas = int(os.getenv("MAX_PRIORITY_FEE_PER_GAS", 0)) * 10**9
        options = {
            "gas_limit": int(os.getenv("GAS_LIMIT", 200000)),
            "max_fee_per_gas": max_fee_per_gas,
            "max_priority_fee_per_gas": max_priority_fee_per_gas,
        }
        return options

    def add_do_request(
        self,
        image_metadata: str,
        payload_metadata: str,
        input_metadata: str,
        node_address: Address,
        resources: dict,
        gas_limit: None = None,
    ) -> HexStr:
        cpu = resources.get("cpu", 1)
        memory = resources.get("memory", 1)
        storage = resources.get("storage", 40)
        bandwidth = resources.get("bandwidth", 1)
        duration = resources.get("duration", 1)
        validators = resources.get("validators", 1)
        task_price = resources.get("task_price", 10)
        tx = self.ethernity_contract.functions._addDORequest(
            cpu,
            memory,
            storage,
            bandwidth,
            duration,
            validators,
            task_price,
            image_metadata,
            payload_metadata,
            input_metadata,
            node_address,
        ).build_transaction(self.__transaction_object)

        signed_tx = self.provider.eth.account.sign_transaction(
            tx, private_key=self.signer._private_key
        )
        self.provider.eth.send_raw_transaction(signed_tx.rawTransaction)
        return self.provider.to_hex(self.provider.keccak(signed_tx.rawTransaction))

    def get_order(self, order_id: int) -> int:
        return self.ethernity_contract.functions._getOrder(order_id).call()

    def approve_order(self, order_id: int) -> HexStr:
        # return self.ethernity_contract.functions._approveOrder(order_id).transact()
        unicorn_txn = self.ethernity_contract.functions._approveOrder(
            order_id
        ).build_transaction(self.__transaction_object)

        signed_txn = self.provider.eth.account.sign_transaction(
            unicorn_txn, private_key=self.signer._private_key
        )
        self.provider.eth.send_raw_transaction(signed_txn.rawTransaction)
        return self.provider.to_hex(self.provider.keccak(signed_txn.rawTransaction))

    def get_result_from_order(self, order_id: int) -> Any:
        self.ethernity_contract.caller()._getOrder(order_id)
        return self.ethernity_contract.caller()._getResultFromOrder(order_id)

    def is_node_operator(self, account: str) -> bool:
        try:
            requests = self.ethernity_contract.functions._getMyDPRequests().call(
                {"from": account}
            )
            return len(requests) > 0
        except Exception as ex:
            print(ex)
            return False

    def sign_message(self, message: str) -> Any:
        if isinstance(message, bytes):
            message = message.decode("utf-8")
        return self.provider.eth.account.sign_message(
            encode_defunct(text=message), self.signer._private_key
        )

    def get_network_name(self) -> Any:
        network = self.provider.eth.chain_id
        return ECNetworkByChainIdDictionary[network]



import base64
import binascii
import codecs
import hashlib
import json
import secrets
from base64 import a85decode, a85encode
from typing import Any

from Crypto.Cipher import AES
from Crypto.PublicKey import ECC

# from cryptography.hazmat.primitives.asymmetric import ec
from nacl.public import Box, PrivateKey, PublicKey
from pyasn1.codec.der import decoder
from tinyec import ec, registry

curve = registry.get_curve("secp384r1")


def compress_point(point: Any) -> str:
    return hex(point.x) + hex(point.y % 2)[2:]


def ecc_point_to_256_bit_key(point: Any) -> bytes:
    value = str(point.x) + str(point.y)
    sha = hashlib.sha256(value.encode())
    return sha.digest()


def encrypt_aes_gcm(msg: bytes, secret_key: bytes) -> tuple:
    aes_cipher = AES.new(secret_key, AES.MODE_GCM)
    cipher_text, auth_tag = aes_cipher.encrypt_and_digest(msg)
    return (cipher_text, aes_cipher.nonce, auth_tag)


def encrypt_ecc(msg: bytes, public_key: Any) -> tuple:
    cipher_text_private_key = secrets.randbelow(curve.field.n)
    shared_ecc_key = cipher_text_private_key * public_key
    secret_key = ecc_point_to_256_bit_key(shared_ecc_key)
    ciphertext, nonce, auth_tag = encrypt_aes_gcm(msg, secret_key)
    cipher_text_public_key = cipher_text_private_key * curve.g
    return (ciphertext, nonce, auth_tag, cipher_text_public_key)


def decrypt_ase_gcm(
    ciphertext: bytes, nonce: bytes, auth_tag: bytes, secret_key: bytes
) -> bytes:
    aes_cipher = AES.new(secret_key, AES.MODE_GCM, nonce)
    plaintext = aes_cipher.decrypt_and_verify(ciphertext, auth_tag)
    return plaintext


def decrypt_ecc(encrypted_msg: tuple, priv_key: int) -> bytes:
    (ciphertext, nonce, authTag, ciphertextPubKey) = encrypted_msg
    shared_ecc_key = priv_key * ciphertextPubKey
    secret_key = ecc_point_to_256_bit_key(shared_ecc_key)
    plaintext = decrypt_ase_gcm(ciphertext, nonce, authTag, secret_key)
    return plaintext


def decrypt_nacl(private_key: str, encrypted_data_hex: str):
    """Decryption function corresponding to encrypt_nacl.

    Args:
        private_key: hex-encoded private key of the recipient (32 bytes).
        encrypted_data_hex: hex string of the encrypted data.

    Returns:
        Decrypted data as bytes.
    """
    try:
        # Convert recipient's private key from hex to PrivateKey object
        recipient_private_key = PrivateKey.from_seed(bytes.fromhex(private_key))

        # Convert the encrypted data from hex to bytes
        encrypted_data = bytes.fromhex(encrypted_data_hex)

        # Extract the ephemeral public key (first 32 bytes)
        emph_public_key_bytes = encrypted_data[:32]
        ciphertext = encrypted_data[32:]

        # Create PublicKey object for the ephemeral public key
        emph_public_key = PublicKey(emph_public_key_bytes)

        # Create a Box for decryption using recipient's private key and ephemeral public key
        dec_box = Box(recipient_private_key, emph_public_key)

        # Decrypt the ciphertext
        decrypted_data = dec_box.decrypt(ciphertext)

        # Decode the data from Ascii85 encoding
        original_data = a85decode(decrypted_data)

        return {"success": True, "data": original_data.decode()}
    except Exception as e:
        print("Decryption error:", str(e))
        return {"success": False, "data": str(e)}


def ecc_read_key(public_key_file: str) -> Any:
    with open(public_key_file) as r:
        key = ECC.import_key(r.read())
        return key


def encrypt(message: bytes, public_key_file: str) -> bytes:
    # with open(public_key_file) as f:
    ecc_public_cert = ECC.import_key(public_key_file)
    pub_key_ecc_point = ec.Point(
        curve,
        ecc_public_cert.pointQ.x.__int__(),
        ecc_public_cert.pointQ.y.__int__(),
    )
    encrypted_msg = encrypt_ecc(message, pub_key_ecc_point)
    return encrypted_data_to_base64_json(encrypted_msg)


def clean_private_key(private_key_data: bytes) -> bytes:
    private_key_data = private_key_data.replace(b"-----BEGIN PRIVATE KEY-----", b"")
    private_key_data = private_key_data.replace(b"-----END PRIVATE KEY-----", b"")
    private_key_data = codecs.decode(private_key_data, "base64")
    return private_key_data


def decrypt(private_key_file: str, encrypted_msg: tuple) -> bytes:
    # Reading and calculating Private Key from PEM
    with open(private_key_file) as f:
        private_key_data = str.encode(f.read())

    private_key_data = clean_private_key(private_key_data)
    # decode der with asn1 library
    # - get the octet string (field-2) containing the raw key
    asn1_object, _ = decoder.decode(private_key_data)
    raw_keys = asn1_object.getComponentByName("field-2").asOctets()
    # - get the octet string (field-1) containing the raw private key
    #   and the bit string (field-2) containing the uncompressed public key
    asn1_object, _ = decoder.decode(raw_keys)
    private_key = asn1_object.getComponentByName("field-1").asOctets()

    # Generating keypair for tinyEC
    priv_key_tec = int.from_bytes(private_key, byteorder="big")
    decrypted_msg = decrypt_ecc(encrypted_msg, priv_key_tec)
    print("\ndecrypted msg:", decrypted_msg)
    return decrypted_msg


def encrypted_data_to_base64_json(encrypted_msg: tuple) -> bytes:
    json_obj = {
        "ciphertext": binascii.hexlify(encrypted_msg[0]).decode(),
        "nonce": binascii.hexlify(encrypted_msg[1]).decode(),
        "authTag": binascii.hexlify(encrypted_msg[2]).decode(),
        "x": hex(encrypted_msg[3].x),
        "y": hex(encrypted_msg[3].y),
    }

    json_string = json.dumps(json_obj)
    binary_data = json_string.encode("utf-8")
    base64_data = base64.b64encode(binary_data)

    return base64_data


def encrypted_data_from_base64_json(base64_data: str) -> tuple:
    # decode the base64-encoded string back to a JSON string
    decoded_json_data = base64.b64decode(base64_data).decode("utf-8")

    json_load = json.loads(decoded_json_data)
    return (
        binascii.unhexlify(json_load["ciphertext"].encode()),
        binascii.unhexlify(json_load["nonce"].encode()),
        binascii.unhexlify(json_load["authTag"].encode()),
        ec.Point(curve, int(json_load["x"], 16), int(json_load["y"], 16)),
    )


def hex_to_bytes(hex_str: str) -> bytes:
    return bytes.fromhex(hex_str[2:] if hex_str[:2] == "0x" else hex_str)


def encrypt_nacl(public_key: str, data: bytes) -> str:
    """Encryption function using NaCl box compatible with MetaMask
    For implementation used in MetaMask look into: https://github.com/MetaMask/eth-sig-util
    Args:
        public_key: public key of recipient (32 bytes)
        data: message data
    Returns:
        encrypted data
    """
    emph_key = PrivateKey.generate()
    enc_box = Box(emph_key, PublicKey(hex_to_bytes(public_key)))
    # Encryption must work with MetaMask decryption (requires valid utf-8)
    data = a85encode(data)
    ciphertext = enc_box.encrypt(data)
    result = bytes(emph_key.public_key) + ciphertext
    return result.hex()


# def compress_point(point: Any) -> str:
#     return hex(point.x) + hex(point.y % 2)[2:]


def sha256(value: str, as_hex: bool = False) -> bytes:
    sha = hashlib.sha256(value.encode()).hexdigest()
    return sha if as_hex else sha.encode()


# def encrypt_with_certificate(message: str, certificate: str) -> str:
#     public_key = load_certificate(certificate)
#     return encrypt(public_key, message)


# def decrypt_with_private_key(
#     encrypted_message: str, account: str
# ) -> dict[str, Any] | dict[str, bool]:
#     try:
#         data = bytes.fromhex(encrypted_message)
#         structured_data = {
#             "version": "x25519-xsalsa20-poly1305",
#             "ephemPublicKey": base64.b64encode(data[:32]).decode(),
#             "nonce": base64.b64encode(data[32:56]).decode(),
#             "ciphertext": base64.b64encode(data[56:]).decode(),
#         }
#         ct = f"0x{base64.b64encode(json.dumps(structured_data).encode()).decode()}"
#         decrypted_message = Web3.eth.account.decrypt(ct, account)
#         decoded_message = decrypted_message.decode("ascii")
#         return {"success": True, "data": decoded_message}
#     except Exception as e:
#         print(e)
#         return {"success": False}


# def encrypt(pub_key: str, message: str) -> str:
#     public_key_point = bytes.fromhex(pub_key)
#     encrypted_message = encrypt_ecc(message, public_key_point)
#     return encrypted_data_to_base64_json(encrypted_message)


# def load_certificate(certificate: str) -> str:
#     cert = x509.load_pem_x509_certificate(certificate.encode(), default_backend())
#     pub_key = cert.public_key()
#     pub_key_bytes = pub_key.public_bytes(
#         encoding=serialization.Encoding.DER,
#         format=serialization.PublicFormat.SubjectPublicKeyInfo,
#     )
#     return pub_key_bytes.hex()


# def encrypt_ecc(message: str, public_key_point_bytes: bytes) -> dict[str, Any]:
#     # Load the public key
#     public_key_point = serialization.load_der_public_key(
#         public_key_point_bytes, backend=default_backend()
#     )

#     # Ensure the public key is on the same curve as the private key
#     if not isinstance(public_key_point.curve, ec.SECP384R1):  # type: ignore
#         raise ValueError("Public key is not on the SECP384R1 curve")

#     # Generate a private key
#     private_key = ec.generate_private_key(ec.SECP384R1(), default_backend())

#     # Compute the shared key
#     shared_key = private_key.exchange(ec.ECDH(), public_key_point)  # type: ignore

#     derived_key = HKDF(
#         algorithm=hashes.SHA256(),
#         length=32,
#         salt=None,
#         info=None,
#         backend=default_backend(),
#     ).derive(shared_key)

#     # Generate a nonce for GCM mode
#     nonce = urandom(12)  # 12 bytes is a common choice for GCM nonce

#     # Initialize the cipher in GCM mode
#     cipher = Cipher(
#         algorithms.AES(derived_key), modes.GCM(nonce), backend=default_backend()
#     )
#     encryptor = cipher.encryptor()

#     # Encrypt the message
#     ciphertext = encryptor.update(message.encode()) + encryptor.finalize()

#     return {
#         "ciphertext": ciphertext,
#         "secretKey": derived_key,
#         "nonce": nonce,
#         "authTag": encryptor.tag,  # The authentication tag generated during encryption
#         "cipherTextPublicKey": private_key.public_key().public_bytes(
#             encoding=serialization.Encoding.PEM,
#             format=serialization.PublicFormat.SubjectPublicKeyInfo,
#         ),
#     }


# def encrypted_data_to_base64_json(encrypted_msg: dict) -> str:
#     public_key = serialization.load_pem_public_key(
#         encrypted_msg["cipherTextPublicKey"], backend=default_backend()
#     )

#     # Ensure the loaded key is an EC public key to access point attributes
#     if not isinstance(public_key, ec.EllipticCurvePublicKey):
#         raise ValueError("The key is not an elliptic curve public key")

#     # Extract the x and y coordinates of the public key
#     public_numbers = public_key.public_numbers()
#     x = public_numbers.x
#     y = public_numbers.y

#     json_obj = {
#         "ciphertext": base64.b64encode(encrypted_msg["ciphertext"]).decode(),
#         "nonce": base64.b64encode(encrypted_msg["nonce"]).decode(),
#         "authTag": base64.b64encode(encrypted_msg["authTag"]).decode(),
#         "x": x.to_bytes((x.bit_length() + 7) // 8, byteorder="big").hex(),
#         "y": y.to_bytes((y.bit_length() + 7) // 8, byteorder="big").hex(),
#     }
#     return base64.b64encode(json.dumps(json_obj).encode()).decode()



contract = {
    "address_bloxberg": "0x15D73a742529C3fb11f3FA32EF7f0CC3870ACA31",
    "address_polygon": "0x689f3806874d3c8A973f419a4eB24e6fBA7E830F",
    "abi": [
        {"inputs": [], "stateMutability": "nonpayable", "type": "constructor"},
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "wallet",
                    "type": "address",
                }
            ],
            "name": "AllowedWalletAdded",
            "type": "event",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "wallet",
                    "type": "address",
                }
            ],
            "name": "AllowedWalletRemoved",
            "type": "event",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": False,
                    "internalType": "string",
                    "name": "ipfsHash",
                    "type": "string",
                },
                {
                    "indexed": False,
                    "internalType": "string",
                    "name": "publicKey",
                    "type": "string",
                },
                {
                    "indexed": False,
                    "internalType": "string",
                    "name": "version",
                    "type": "string",
                },
            ],
            "name": "ImageAdded",
            "type": "event",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": False,
                    "internalType": "string",
                    "name": "ipfsHash",
                    "type": "string",
                }
            ],
            "name": "ImageRemoved",
            "type": "event",
        },
        {
            "inputs": [
                {"internalType": "address", "name": "wallet", "type": "address"}
            ],
            "name": "addAllowedWallet",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "string", "name": "ipfsHash", "type": "string"},
                {"internalType": "string", "name": "certPublicKey", "type": "string"},
                {"internalType": "string", "name": "version", "type": "string"},
                {"internalType": "string", "name": "imageName", "type": "string"},
                {
                    "internalType": "string",
                    "name": "dockerComposeHash",
                    "type": "string",
                },
            ],
            "name": "addImage",
            "outputs": [],
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "string", "name": "imageName", "type": "string"},
                {"internalType": "string", "name": "version", "type": "string"},
            ],
            "name": "getLatestTrustedZoneImageCertPublicKey",
            "outputs": [
                {"internalType": "string", "name": "", "type": "string"},
                {"internalType": "string", "name": "", "type": "string"},
                {"internalType": "string", "name": "", "type": "string"},
            ],
            "stateMutability": "view",
            "type": "function",
        },
        {
            "inputs": [
                {"internalType": "string", "name": "imageName", "type": "string"},
                {"internalType": "string", "name": "version", "type": "string"},
            ],
            "name": "getLatestImageVersionPublicKey",
            "outputs": [
                {"internalType": "string", "name": "", "type": "string"},
                {"internalType": "string", "name": "", "type": "string"},
                {"internalType": "string", "name": "", "type": "string"},
            ],
            "stateMutability": "view",
            "type": "function",
        },
    ],
}



import os
import time
from datetime import datetime
from typing import Any, Literal, Union

from eth_account import Account
from eth_typing import Address, HexStr
from web3 import Web3
from web3.contract.contract import Contract
from web3.exceptions import TimeExhausted, TransactionNotFound

from .contract.abi.bloxbergAbi import contract as bloxbergAbi
from .contract.abi.polygonAbi import contract as polygonAbi
from .contract.operation.bloxbergProtocolContract import BloxbergProtocolContract
from .contract.operation.imageRegistryContract import ImageRegistryContract
from .contract.operation.polygonProtocolContract import PolygonProtocolContract
from binascii import hexlify
from nacl.public import Box, PrivateKey, PublicKey

# from .crypto import decrypt_with_private_key, encrypt_with_certificate, sha256
from .crypto import decrypt_nacl, encrypt, sha256
from .enums import (
    ZERO_CHECKSUM,
    ECAddress,
    ECError,
    ECEvent,
    ECOrderTaskStatus,
    ECStatus,
    ECNetworkEnvToEnum,
    ECLog,
)
from .ipfs import IPFSClient
from .utils import (
    format_date,
    generate_random_hex_of_size,
    is_address,
    is_null_or_empty,
    parse_transaction_bytes_ut,
    generate_wallet,
)

try:
    from dotenv import load_dotenv

    load_dotenv(".env" if os.path.exists(".env") else ".env.config")
except ImportError as e:
    pass


ipfs_address = os.environ.get("IPFS_ADDRESS", "")
if not ipfs_address:
    ipfs_address = "https://ipfs.ethernity.cloud/api/v0"
ipfs_token = os.environ.get("IPFS_TOKEN", "")

LAST_BLOCKS = 20
VERSION = "v3"
NETWORK = os.environ.get("BLOCKCHAIN_NETWORK", None)

if len(os.environ.get("PRIVATE_KEY", "")) < 10:
    raise Exception("PRIVATE_KEY is not set in .env file")

SIGNER_ACCOUNT = Account().from_key(os.environ.get("PRIVATE_KEY"))

ipfsClient = Any


class EthernityCloudRunner:
    def __init__(self, network_address: Address = ECAddress.BLOXBERG.TESTNET_ADDRESS) -> None:  # type: ignore
        self.node_address = ""
        self.challenge_hash = ""
        self.order_number = -1
        self.do_hash = None
        self.do_request_id = -1
        self.do_request = -1
        self.script_hash = ""
        self.file_set_hash = ""
        self.interval = None
        self.order_placed_timer = None
        self.task_has_been_picked_for_approval = False
        self.get_result_from_order_repeats = 1
        self.network_address = network_address
        self.enclave_image_ipfs_hash = ""
        self.enclave_public_key = ""
        self.enclave_docker_compose_ipfs_hash = ""
        self.trustedZoneImage = ""
        self.log_level = ECLog.ERROR
        
        # for configurations with contract addresses
        if NETWORK is not None:
            network_address = ECNetworkEnvToEnum.get(
                NETWORK.lower(), ECAddress.BLOXBERG.TESTNET_ADDRESS
            )

        if network_address in [
            ECAddress.BLOXBERG.TESTNET_ADDRESS,
            ECAddress.BLOXBERG.MAINNET_ADDRESS,
        ]:
            self.protocol_contract = BloxbergProtocolContract(
                network_address, SIGNER_ACCOUNT
            )
            self.protocol_abi = bloxbergAbi.get("abi")
        elif network_address == ECAddress.POLYGON.MAINNET_ADDRESS:
            self.protocol_contract = PolygonProtocolContract(
                ECAddress.POLYGON.MAINNET_PROTOCOL_ADDRESS, SIGNER_ACCOUNT, True  # type: ignore
            )
            self.protocol_abi = polygonAbi.get("abi")
        elif network_address == ECAddress.POLYGON.TESTNET_ADDRESS:
            self.protocol_contract = PolygonProtocolContract(
                ECAddress.POLYGON.TESTNET_PROTOCOL_ADDRESS, SIGNER_ACCOUNT, False  # type: ignore
            )
            self.protocol_abi = polygonAbi.get("abi")

        self.token_contract = self.protocol_contract.ethernity_contract

        self.initialize_storage(ipfs_address, ipfs_token)

    def set_log_level(self, level):
        self.log_level = ECLog[level.upper()]

    def is_mainnet(self) -> bool:
        return self.network_address in [
            ECAddress.BLOXBERG.MAINNET_ADDRESS,
            ECAddress.POLYGON.MAINNET_ADDRESS,
        ]

    def dispatch_ec_event(
        self,
        message: Any,
        status: str = ECStatus.DEFAULT,
        event_type: str = ECEvent.TASK_PROGRESS,
        log_level: str = ECLog.ERROR,
    ) -> None:
        # Custom event dispatching logic
        if (self.log_level >= log_level):
            debug_details = ""
            if (self.log_level == ECLog.DEBUG):
                debug_details = "Event:",{event_type},"Status:",{status}
            print(f"[{self.log_level}] {datetime.now()} {message}", debug_details)
                  

    def get_enclave_details(self) -> None:
        details = self.image_registry_contract.get_enclave_details_v3(
            self.runner_type, VERSION, self.trustedZoneImage
        )
        if details:
            (
                self.enclave_image_ipfs_hash,
                self.enclave_public_key,
                self.enclave_docker_compose_ipfs_hash,
            ) = details
            self.dispatch_ec_event(
                f"ENCLAVE_IMAGE_IPFS_HASH: {self.enclave_image_ipfs_hash}",ECStatus.DEFAULT,ECEvent.TASK_NEW,ECLog.INFO
            )
            self.dispatch_ec_event(f"ENCLAVE_PUBLIC_KEY:\n{self.enclave_public_key}",ECStatus.DEFAULT,ECEvent.TASK_NEW,ECLog.DEBUG)
            self.dispatch_ec_event(
                f"ENCLAVE_DOCKER_COMPOSE_IPFS_HASH: {self.enclave_docker_compose_ipfs_hash}",ECStatus.DEFAULT,ECEvent.TASK_NEW,ECLog.INFO
            )

    def get_reason(
        self, contract: Contract, tx_hash: str
    ) -> Literal["Transaction hash not found"]:
        tx = contract.get_provider().get_transaction(tx_hash)  # type: ignore
        if not tx:
            return "Transaction not found"
        del tx.gas_price
        code = contract.get_provider().call(tx, tx.block_number)  # type: ignore
        reason = Web3.to_utf8_string(f"0x{code[138:]}")  # type: ignore
        #print(reason)
        return reason.strip()

    def wait_for_transaction_to_be_processed(
        self, contract: Contract, transaction_hash: str, attempt: int = 0
    ) -> bool:
        while True and attempt < 10:
            try:
                contract.get_provider().eth.wait_for_transaction_receipt(transaction_hash)  # type: ignore
                tx_receipt = contract.get_provider().eth.get_transaction_receipt(transaction_hash)  # type: ignore
                if tx_receipt:
                    break
            except TransactionNotFound or TimeExhausted:
                # we need to sleep here to avoid spamming the node
                time.sleep(3)
                self.wait_for_transaction_to_be_processed(
                    contract, transaction_hash, attempt + 1
                )

        #print(tx_receipt)
        return not (not tx_receipt and tx_receipt.status == 0)

    def check_web3_connection(self) -> Literal[False]:
        try:
            # self.token_contract.get_provider().send("eth_requestAccounts", [])
            self.protocol_contract.get_provider()
            wallet_address = self.protocol_contract.get_current_wallet()
            return wallet_address is not None and wallet_address != ""
        except Exception as e:
            self.dispatch_ec_event(
                f"{e}",ECStatus.ERROR,ECEvent.TASK_CREATED,ECLog.ERROR
            )
            return False

    def hex_to_bytes(self, hex_str: str) -> bytes:
        return bytes.fromhex(hex_str[2:] if hex_str[:2] == "0x" else hex_str)

    def bytes_to_hex(self, bytes_str: bytes) -> str:
        return "0x" + hexlify(bytes_str).decode("utf-8")

    def get_current_wallet_public_key(self) -> HexStr:
        return self.bytes_to_hex(
            PrivateKey.from_seed(
                self.hex_to_bytes(os.environ.get("PRIVATE_KEY", ""))
            ).public_key._public_key
        )

    def get_v3_image_metadata(self, challenge_hash: str) -> str:
        base64_encrypted_challenge = encrypt(challenge_hash, self.enclave_public_key)
        challenge_ipfs_hash = ipfsClient.upload_to_ipfs(base64_encrypted_challenge)
        if challenge_ipfs_hash != None:
            self.dispatch_ec_event(
                f"Uploaded challenge to IPFS: {challenge_ipfs_hash}",ECStatus.DEFAULT,ECEvent.TASK_NEW,ECLog.INFO
            )
        public_key = self.get_current_wallet_public_key()
        return f"{VERSION}:{self.enclave_image_ipfs_hash}:{self.runner_type if not self.trustedZoneImage else self.trustedZoneImage}:{self.enclave_docker_compose_ipfs_hash}:{challenge_ipfs_hash}:{public_key}"

    def get_v3_code_metadata(self, code: str) -> str:
        script_checksum = sha256(code)
        base64_encrypted_script = encrypt(code.encode("utf-8"), self.enclave_public_key)
        self.script_hash = ipfsClient.upload_to_ipfs(base64_encrypted_script)
        if self.script_hash!= None:
            self.dispatch_ec_event(
                f"Uploaded encrypted code to IPFS: {self.script_hash}",ECStatus.DEFAULT,ECEvent.TASK_NEW,ECLog.INFO
            )
        script_checksum = self.protocol_contract.sign_message(script_checksum)
        return f"{VERSION}:{self.script_hash}:{script_checksum.signature.hex()}"

    def get_v3_input_metadata(self) -> str:
        file_set_checksum = self.protocol_contract.sign_message(ZERO_CHECKSUM)
        return f"{VERSION}::{file_set_checksum.signature.hex()}"

    def create_do_request(
        self,
        image_metadata: str,
        code_metadata: str,
        input_metadata: str,
        node_address: Address,
        gas_limit: None = None,
    ) -> bool:
        try:
            __provider = self.protocol_contract.get_provider()
            self.dispatch_ec_event(
                f"Submitting transaction for DO request",ECStatus.DEFAULT,ECEvent.TASK_NEW,ECLog.INFO
            )
            transaction_hash = self.protocol_contract.add_do_request(
                image_metadata,
                code_metadata,
                input_metadata,
                node_address,
                self.resources,
            )

            self.do_hash = transaction_hash
            receipt = None
            for i in range(100):
                try:
                    self.dispatch_ec_event(
                        f"{transaction_hash} is pending...",ECStatus.DEFAULT,ECEvent.TASK_NEW,ECLog.INFO
                    )

                    is_processed = self.wait_for_transaction_to_be_processed(
                        self.protocol_contract, transaction_hash
                    )
                    receipt = __provider.eth.wait_for_transaction_receipt(
                        transaction_hash
                    )
                    processed_logs = (
                        self.token_contract.events._addDORequestEV().process_receipt(
                            receipt
                        )
                    )
                    self.do_request = processed_logs[0].args._rowNumber
                except KeyError:
                    time.sleep(1)
                    continue
                except Exception as e:
                    self.dispatch_ec_event(
                        f"{e}",ECStatus.ERROR,ECEvent.TASK_CREATED,ECLog.ERROR
                    )
                    raise
                else:
                    self.dispatch_ec_event(
                        f"{transaction_hash} confirmed!",ECStatus.ERROR,ECEvent.TASK_CREATED,ECLog.INFO
                    )
                    self.dispatch_ec_event(
                        f"Request {self.do_request} created successfully!",ECStatus.DEFAULT,ECEvent.TASK_CREATED,ECLog.INFO
                    )
                    self.do_hash = transaction_hash
                    break

            if not is_processed:
                reason = self.get_reason(self.protocol_contract, transaction_hash)
                self.dispatch_ec_event(
                        f"Unable to create DO request: {reason}",ECStatus.ERROR,ECEvent.TASK_NEW,ECLog.ERROR
                )
                return False

            return True
        except Exception as e:
            self.dispatch_ec_event(
                f"{e}",ECStatus.ERROR,ECEvent.TASK_NEW,ECLog.ERROR
            )
            # TODO Calculate gas propely
            if (
                "cannot estimate gas; transaction may fail or may require manual gas limit"
                in str(e)
            ):
                return self.create_do_request(
                    image_metadata, code_metadata, input_metadata, node_address, 5000000
                )
            return False

    def parse_order_result(self, result: str) -> dict[str, Any]:
        try:
            arr = result.split(":")
            t_bytes = arr[1] if arr[1].startswith("0x") else f"0x{arr[1]}"
            return {
                "version": arr[0],
                "transaction_bytes": t_bytes,
                "result_ipfs_hash": arr[2],
            }
        except Exception:
            raise ValueError(ECError.PARSE_ERROR)

    def parse_transaction_bytes(self, bytes: bytes) -> dict[str, str]:
        try:
            result = parse_transaction_bytes_ut(self.protocol_abi, bytes)
            arr = result["result"].split(":")
            return {
                "version": arr[0],
                "from": result["from"],
                "task_code": arr[1],
                "task_code_string": ECOrderTaskStatus[int(arr[1])],
                "checksum": arr[2],
                "enclave_challenge": arr[3],
            }
        except Exception as ex:
            raise ValueError(ECError.PARSE_ERROR)

    def get_result_from_order(self, order_id: int) -> Any:
        decrypted_data = {}
        try:
            # stuck here
            order_result = self.protocol_contract.get_result_from_order(order_id)
            # self.protocol_contract.ethernity_contract.caller(
            #     transaction={"from": SIGNER_ACCOUNT.address}
            # )._getResultFromOrder(order_id)
            self.dispatch_ec_event(
                f"Task {order_id} was successfully processed.",ECStatus.SUCCESS,ECEvent.TASK_COMPLETED,ECLog.INFO
            )
            parsed_order_result = self.parse_order_result(order_result)
            self.dispatch_ec_event(
                f"Result IPFS hash: {parsed_order_result['result_ipfs_hash']}",ECStatus.SUCCESS,ECEvent.TASK_COMPLETED,ECLog.INFO
            }

            transaction_result = self.parse_transaction_bytes(
                parsed_order_result["transaction_bytes"]
            )

            wallet = generate_wallet(
                self.challenge_hash.decode("utf-8"),  # type: ignore
                transaction_result["enclave_challenge"],
            )
            if not wallet or wallet != transaction_result["from"]:
                return {
                    "success": False,
                    "message": "Integrity check failed, signer wallet address is wrong.",
                }
            self.dispatch_ec_event(
                f"Downloading {parsed_order_result['result_ipfs_hash']} ...",ECStatus.SUCCESS,ECEvent.TASK_COMPLETED,ECLog.INFO
            )
            ipfs_result = ipfsClient.get_file_content(
                parsed_order_result["result_ipfs_hash"]
            )
            self.dispatch_ec_event(
                f"Validating proof result...",ECStatus.SUCCESS,ECEvent.TASK_COMPLETED,ECLog.INFO
            )
            current_wallet_address = self.protocol_contract.get_current_wallet()
            # decrypted_data = decrypt_with_private_key(
            decrypted_data = decrypt_nacl(os.environ.get("PRIVATE_KEY"), ipfs_result)
            if not decrypted_data["success"]:
                return {
                    "success": False,
                    "message": "Could not decrypt the order result.",
                }

            self.dispatch_ec_event(f"Result value: {decrypted_data['data']}",ECStatus.SUCCESS,ECEvent.TASK_COMPLETED,ECLog.DEBUG)
            ipfs_result_checksum = sha256(decrypted_data["data"],True)
            if ipfs_result_checksum != transaction_result["checksum"]:
                return {
                    "success": False,
                    "message": "Integrity check failed, checksum of the order result is wrong. {ipfs_result_checksum} != {transaction_result['checksum']}",
                }
            
            self.dispatch_ec_event(
                f"Collecting result transaction hash...",ECStatus.SUCCESS,ECEvent.TASK_COMPLETED,ECLog.INFO
            )
            transaction = self.protocol_contract.get_provider().eth.get_transaction_receipt(
                self.do_hash
            )
            block = self.protocol_contract.get_provider().eth.get_block(
                transaction.blockNumber
            )
            block_timestamp = block.timestamp
            end_block_number = self.protocol_contract.get_provider().eth.get_block_number()
            start_block_number = end_block_number - LAST_BLOCKS
            result_transaction_hash = None
            result_block_timestamp = None
            for i in range(end_block_number, start_block_number - 1, -1):
                block = (
                    self.protocol_contract.get_provider().eth.get_block(i)
                )
                if not block or not block.transactions:
                    continue
                # TODO: fix output transaction reporting
                for transaction in block.transactions:
                    transaction_details = self.protocol_contract.get_provider().eth.get_transaction(transaction)
                    if (
                        transaction_details.to == self.protocol_contract.contract_address()
                    ):
                        result_transaction_hash = transaction
                        result_block_timestamp = block.timestamp
            return {
                "success": True,
                "contract_address": self.protocol_contract.contract_address(),
                "input_transaction_hash": self.do_hash,
                "output_transaction_hash": result_transaction_hash,
                "order_id": order_id,
                "image_hash": f"{self.enclave_image_ipfs_hash}:{self.runner_type}",
                "script_hash": self.script_hash,
                "file_set_hash": self.file_set_hash,
                "public_timestamp": block_timestamp,
                "result_hash": parsed_order_result["result_ipfs_hash"],
                "result_task_code": transaction_result["task_code_string"],
                "result_value": ipfs_result,
                "result_timestamp": result_block_timestamp,
                "result": decrypted_data["data"],
            }
        except Exception as e:
            self.dispatch_ec_event(
                f"{e}",ECStatus.ERROR,ECEvent.TASK_PROGRESS,ECLog.ERROR
            )
            if str(e) == ECError.PARSE_ERROR:
                return {
                    "success": False,
                    "message": "Ethernity parsing transaction error.",
                }
            if str(e) == ECError.IPFS_DOWNLOAD_ERROR:
                return {
                    "success": False,
                    "message": "Ethernity IPFS download result error.",
                }
            time.sleep(5)
            self.get_result_from_order_repeats += 1
            return self.get_result_from_order(order_id)

    def process_task(self, code: str) -> bool:
        # self.listen_for_add_do_request_event()
        self.challenge_hash = generate_random_hex_of_size(20).encode("utf-8")
        image_metadata = self.get_v3_image_metadata(self.challenge_hash)
        code_metadata = self.get_v3_code_metadata(code)
        input_metadata = self.get_v3_input_metadata()
        do_sent_successfully = self.create_do_request(
            image_metadata, code_metadata, input_metadata, self.node_address, None
        )
        if do_sent_successfully:
            result = self._wait_for_processor()
            return result

        return False

    def _wait_for_processor(self) -> None:
        self.dispatch_ec_event(
            f"Waiting for Ethernity CLOUD network...",ECStatus.DEFAULT,ECEvent.TASK_CREATED,ECLog.INFO
        )

        while True:
            try:
                order = self.find_order(self.do_request)
            except Exception as e:
                self.dispatch_ec_event(
                    f"{e}",ECStatus.ERROR,ECEvent.TASK_CREATED,ECLog.ERROR
                )
            if order is not None:
                self.dispatch_ec_event(
                    f"Connected !",ECStatus.DEFAULT,ECEvent.TASK_ORDER_PLACED,ECLog.INFO
                )
                self.dispatch_ec_event(
                    f"Approving task {order}...",ECStatus.DEFAULT,ECEvent.TASK_ORDER_PLACED,ECLog.INFO
                )
                if self.approve_order(order):
                    break

                result = self.get_result_from_order(order)

                if result:
                    return result
                
                break
                time.sleep(5)
            else:
                time.sleep(5)

    def find_order(self, doreq: int) -> Union[int, None]:
        count = self.token_contract.functions._getOrdersCount().call()
        for i in range(count - 1, count - 5, -1):
            order = self.token_contract.caller()._getOrder(i)
            # if order[2] == doreq and order[4] == 0:
            if order[2] == doreq:
                return i
        return None

    def approve_order(self, order: int) -> bool:
        transaction_hash = self.protocol_contract.approve_order(order)

        receipt = None
        for i in range(100):
            try:
                self.dispatch_ec_event(
                        f"{transaction_hash} is pending...",ECStatus.DEFAULT,ECEvent.TASK_NEW,ECLog.INFO
                )
                receipt = self.protocol_contract.get_provider().eth.wait_for_transaction_receipt(
                    transaction_hash
                )
                self.node_address = self.protocol_contract.get_order(order)[1]
            except KeyError:
                time.sleep(1)
                continue
            except TimeExhausted:
                raise
            except Exception as e:
                self.dispatch_ec_event(
                    f"{e}",ECStatus.ERROR,ECEvent.TASK_ORDER_PLACED,ECLog.ERROR
                )
                raise
            else:
                self.dispatch_ec_event(
                    f"{transaction_hash} confirmed!",ECStatus.DEFAULT,ECEvent.TASK_PROGRESS,ECLog.INFO
                )
                self.dispatch_ec_event(
                    f"Task {order} approved successfully!",ECStatus.DEFAULT,ECEvent.TASK_PROGRESS,ECLog.INFO
                )

                break

        if receipt is None:
            self.dispatch_ec_event(
                f"Unable to approve order, please check connectivity with RPC endpoint!",ECStatus.DEFAULT,ECEvent.TASK_PROGRESS,ECLog.WARNING
            )
            return True

        return False

    def reset(self) -> None:
        self.order_number = -1
        self.do_hash = None
        self.do_request_id = -1
        self.do_request = -1
        self.script_hash = ""
        self.file_set_hash = ""
        self.interval = None
        self.get_result_from_order_repeats = 1
        self.task_has_been_picked_for_approval = False

    def cleanup(self) -> None:
        self.reset()
        contract = self.protocol_contract.get_contract()
        # contract.remove_all_listeners()

    def is_node_operator_address(self, node_address: str) -> bool:
        if is_null_or_empty(node_address):
            
            return True
        if is_address(node_address):
            is_node = self.protocol_contract.is_node_operator(node_address)
            if not is_node:
                self.dispatch_ec_event(
                    "The target address is not a valid node operator address",
                    ECStatus.ERROR,ECEvent.TASK_NEW,ECLog.ERROR
                )
                return False
            return True
        self.dispatch_ec_event(
            "Introduced address is not a valid wallet address.", ECStatus.ERROR,ECEvent.TASK_NEW,ECLog.ERROR
        )
        return False

    def initialize_storage(self, ipfs_address: str, token: str = "") -> None:
        global ipfsClient
        ipfsClient = IPFSClient(ipfs_address, token)

    def run(
        self,
        runner_type: str,
        code: str,
        node_address: str,
        resources: Union[dict, None] = None,
        trustedZoneImage: str = "",
    ) -> None:
        if resources is None:
            self.resources = {
                "taskPrice": 10,
                "cpu": 1,
                "memory": 1,
                "storage": 40,
                "bandwidth": 1,
                "duration": 1,
                "validators": 1,
            }
            resources = self.resources
        else:
            self.resources = resources
        if trustedZoneImage:
            self.trustedZoneImage = trustedZoneImage
        try:
            balance = self.protocol_contract.get_balance()
            balance = int(balance)

            if balance < resources["taskPrice"]:
                self.dispatch_ec_event(
                    f"Your wallet balance is insufficient to cover the requested task price. Required: {resources['taskPrice']}, Availabie: {balance}",
                    ECStatus.ERROR,ECEvent.TASK_NEW,ECLog.ERROR
                )
                return

            self.node_address = node_address

            is_node_operator_address = self.is_node_operator_address(node_address)
            if is_node_operator_address:
                self.runner_type = runner_type
                self.cleanup()
                self.dispatch_ec_event("Started processing task...",ECStatus.DEFAULT,ECEvent.TASK_NEW,ECLog.INFO)
                self.image_registry_contract = ImageRegistryContract(
                    self.network_address,
                    (
                        self.runner_type
                        if not self.trustedZoneImage
                        else self.trustedZoneImage
                    ),
                    SIGNER_ACCOUNT,
                )
                connected = self.check_web3_connection()
                if connected:
                    self.get_enclave_details()
                    if self.network_address in [
                        ECAddress.POLYGON.MAINNET_ADDRESS,
                        ECAddress.POLYGON.TESTNET_ADDRESS,
                    ]:
                        self.dispatch_ec_event(
                            "Checking for the allowance on the current wallet...",ECStatus.DEFAULT,ECEvent.TASK_NEW,ECLog.INFO
                        )
                        passed_allowance = self.token_contract.check_and_set_allowance(
                            self.protocol_contract.contract_address(),
                            "100",
                            str(resources["taskPrice"]),
                        )
                        if not passed_allowance:
                            self.dispatch_ec_event(
                                "Unable to set allowance.", ECStatus.ERROR,ECEvent.TASK_NEW,ECLog.ERROR
                            )
                            return
                        self.dispatch_ec_event("Allowance checking completed.",ECStatus.DEFAULT,ECEvent.TASK_NEW,ECLog.INFO)

                    result = self.process_task(code)
                    if not result:
                        self.dispatch_ec_event(
                            "Unable to proceed with the task; exiting.", ECStatus.ERROR,ECEvent.TASK_NEW,ECLog.ERROR
                        )
                        return False
                    return result
                else:
                    self.dispatch_ec_event(
                        "Unable to connect to RPC endpoint", ECStatus.ERROR,ECEvent.TASK_NEW,ECLog.ERROR
                    )
        except Exception as e:
            self.dispatch_ec_event(e, ECStatus.ERROR,ECEvent.TASK_NEW,ECLog.ERROR)



import os

contract = {
    "address": os.getenv(
        "ETNY_CONTRACT_ADDRESS", "0x02882F03097fE8cD31afbdFbB5D72a498B41112c"
    ),
    "abi": [
        {
            "constant": False,
            "inputs": [
                {"internalType": "uint8", "name": "_cpuRequest", "type": "uint8"},
                {"internalType": "uint8", "name": "_memRequest", "type": "uint8"},
                {"internalType": "uint8", "name": "_storageRequest", "type": "uint8"},
                {"internalType": "uint8", "name": "_bandwidthRequest", "type": "uint8"},
                {"internalType": "uint16", "name": "_duration", "type": "uint16"},
                {"internalType": "uint8", "name": "_instances", "type": "uint8"},
                {"internalType": "uint8", "name": "_maxPrice", "type": "uint8"},
                {"internalType": "string", "name": "_metadata1", "type": "string"},
                {"internalType": "string", "name": "_metadata2", "type": "string"},
                {"internalType": "string", "name": "_metadata3", "type": "string"},
                {"internalType": "string", "name": "_metadata4", "type": "string"},
            ],
            "name": "_addDORequest",
            "outputs": [
                {"internalType": "uint256", "name": "_rowNumber", "type": "uint256"}
            ],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "constant": False,
            "inputs": [
                {"internalType": "uint8", "name": "_cpuRequest", "type": "uint8"},
                {"internalType": "uint8", "name": "_memRequest", "type": "uint8"},
                {"internalType": "uint8", "name": "_storageRequest", "type": "uint8"},
                {"internalType": "uint8", "name": "_bandwidthRequest", "type": "uint8"},
                {"internalType": "uint16", "name": "_duration", "type": "uint16"},
                {"internalType": "uint8", "name": "_minPrice", "type": "uint8"},
                {"internalType": "string", "name": "_metadata1", "type": "string"},
                {"internalType": "string", "name": "_metadata2", "type": "string"},
                {"internalType": "string", "name": "_metadata3", "type": "string"},
                {"internalType": "string", "name": "_metadata4", "type": "string"},
            ],
            "name": "_addDPRequest",
            "outputs": [
                {"internalType": "uint256", "name": "_rowNumber", "type": "uint256"}
            ],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "constant": False,
            "inputs": [
                {
                    "internalType": "uint256",
                    "name": "_doRequestItem",
                    "type": "uint256",
                },
                {
                    "internalType": "uint256",
                    "name": "_dpRequestItem",
                    "type": "uint256",
                },
            ],
            "name": "_placeOrder",
            "outputs": [
                {"internalType": "uint256", "name": "_orderNumber", "type": "uint256"}
            ],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "tokenOwner",
                    "type": "address",
                },
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "spender",
                    "type": "address",
                },
                {
                    "indexed": False,
                    "internalType": "uint256",
                    "name": "tokens",
                    "type": "uint256",
                },
            ],
            "name": "Approval",
            "type": "event",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "_from",
                    "type": "address",
                },
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "_to",
                    "type": "address",
                },
            ],
            "name": "OwnershipTransferred",
            "type": "event",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "_from",
                    "type": "address",
                },
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "_to",
                    "type": "address",
                },
            ],
            "name": "ProxyTransferred",
            "type": "event",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "from",
                    "type": "address",
                },
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "to",
                    "type": "address",
                },
                {
                    "indexed": False,
                    "internalType": "uint256",
                    "name": "tokens",
                    "type": "uint256",
                },
            ],
            "name": "Transfer",
            "type": "event",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "_from",
                    "type": "address",
                },
                {
                    "indexed": False,
                    "internalType": "uint256",
                    "name": "_rowNumber",
                    "type": "uint256",
                },
            ],
            "name": "_addDORequestEV",
            "type": "event",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "_from",
                    "type": "address",
                },
                {
                    "indexed": False,
                    "internalType": "uint256",
                    "name": "_rowNumber",
                    "type": "uint256",
                },
            ],
            "name": "_addDPRequestEV",
            "type": "event",
        },
        {
            "constant": False,
            "inputs": [
                {
                    "internalType": "uint256",
                    "name": "_requestListItem",
                    "type": "uint256",
                },
                {"internalType": "string", "name": "_key", "type": "string"},
                {"internalType": "string", "name": "_value", "type": "string"},
            ],
            "name": "_addMetadataToDPRequest",
            "outputs": [
                {"internalType": "uint256", "name": "_rowNumber", "type": "uint256"}
            ],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "constant": False,
            "inputs": [
                {
                    "internalType": "uint256",
                    "name": "_requestListItem",
                    "type": "uint256",
                },
                {"internalType": "string", "name": "_key", "type": "string"},
                {"internalType": "string", "name": "_value", "type": "string"},
            ],
            "name": "_addMetadataToRequest",
            "outputs": [
                {"internalType": "uint256", "name": "_rowNumber", "type": "uint256"}
            ],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "constant": False,
            "inputs": [
                {"internalType": "uint256", "name": "_orderItem", "type": "uint256"},
                {"internalType": "address", "name": "processor", "type": "address"},
            ],
            "name": "_addProcessorToOrder",
            "outputs": [{"internalType": "bool", "name": "success", "type": "bool"}],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "constant": False,
            "inputs": [
                {"internalType": "uint256", "name": "_orderItem", "type": "uint256"},
                {"internalType": "string", "name": "_result", "type": "string"},
            ],
            "name": "_addResultToOrder",
            "outputs": [{"internalType": "bool", "name": "success", "type": "bool"}],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "constant": False,
            "inputs": [{"internalType": "address", "name": "addr", "type": "address"}],
            "name": "_addToPresaleRound",
            "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "constant": False,
            "inputs": [{"internalType": "address", "name": "addr", "type": "address"}],
            "name": "_addToPrivateSaleRound",
            "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "constant": False,
            "inputs": [{"internalType": "address", "name": "addr", "type": "address"}],
            "name": "_addToPublicOneRound",
            "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "constant": False,
            "inputs": [{"internalType": "address", "name": "addr", "type": "address"}],
            "name": "_addToPublicTwoRound",
            "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "constant": False,
            "inputs": [{"internalType": "address", "name": "addr", "type": "address"}],
            "name": "_addToSeedRound",
            "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "constant": False,
            "inputs": [
                {"internalType": "uint256", "name": "_orderItem", "type": "uint256"}
            ],
            "name": "_approveOrder",
            "outputs": [{"internalType": "bool", "name": "success", "type": "bool"}],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "constant": False,
            "inputs": [
                {
                    "internalType": "uint256",
                    "name": "_requestListItem",
                    "type": "uint256",
                }
            ],
            "name": "_cancelDORequest",
            "outputs": [],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "constant": False,
            "inputs": [
                {
                    "internalType": "uint256",
                    "name": "_requestListItem",
                    "type": "uint256",
                }
            ],
            "name": "_cancelDPRequest",
            "outputs": [],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "_downer",
                    "type": "address",
                },
                {
                    "indexed": False,
                    "internalType": "address",
                    "name": "_dproc",
                    "type": "address",
                },
                {
                    "indexed": False,
                    "internalType": "uint256",
                    "name": "_orderNumber",
                    "type": "uint256",
                },
            ],
            "name": "_orderApprovedEV",
            "type": "event",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": False,
                    "internalType": "uint256",
                    "name": "_orderNumber",
                    "type": "uint256",
                }
            ],
            "name": "_orderClosedEV",
            "type": "event",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": False,
                    "internalType": "uint256",
                    "name": "_orderNumber",
                    "type": "uint256",
                },
                {
                    "indexed": False,
                    "internalType": "uint256",
                    "name": "_doRequestId",
                    "type": "uint256",
                },
                {
                    "indexed": False,
                    "internalType": "uint256",
                    "name": "_dpRequestId",
                    "type": "uint256",
                },
            ],
            "name": "_orderPlacedEV",
            "type": "event",
        },
        {
            "anonymous": False,
            "inputs": [
                {
                    "indexed": True,
                    "internalType": "address",
                    "name": "_from",
                    "type": "address",
                },
                {
                    "indexed": False,
                    "internalType": "uint256",
                    "name": "_orderNumber",
                    "type": "uint256",
                },
            ],
            "name": "_placeOrderEV",
            "type": "event",
        },
        {
            "constant": False,
            "inputs": [{"internalType": "address", "name": "addr", "type": "address"}],
            "name": "_removeLockoutTimestamp",
            "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "constant": False,
            "inputs": [
                {"internalType": "uint128", "name": "timestamp", "type": "uint128"}
            ],
            "name": "_setBaseLockTimestamp",
            "outputs": [{"internalType": "bool", "name": "", "type": "bool"}],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "constant": False,
            "inputs": [],
            "name": "acceptOwnership",
            "outputs": [],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "constant": False,
            "inputs": [
                {"internalType": "address", "name": "spender", "type": "address"},
                {"internalType": "uint256", "name": "tokens", "type": "uint256"},
            ],
            "name": "approve",
            "outputs": [{"internalType": "bool", "name": "success", "type": "bool"}],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "constant": False,
            "inputs": [
                {"internalType": "address", "name": "spender", "type": "address"},
                {"internalType": "uint256", "name": "tokens", "type": "uint256"},
                {"internalType": "bytes", "name": "data", "type": "bytes"},
            ],
            "name": "approveAndCall",
            "outputs": [{"internalType": "bool", "name": "success", "type": "bool"}],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "constant": False,
            "inputs": [],
            "name": "faucet",
            "outputs": [{"internalType": "bool", "name": "success", "type": "bool"}],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "constant": False,
            "inputs": [
                {"internalType": "address", "name": "to", "type": "address"},
                {"internalType": "uint256", "name": "tokens", "type": "uint256"},
            ],
            "name": "transfer",
            "outputs": [{"internalType": "bool", "name": "success", "type": "bool"}],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "constant": False,
            "inputs": [
                {"internalType": "address", "name": "from", "type": "address"},
                {"internalType": "address", "name": "to", "type": "address"},
                {"internalType": "uint256", "name": "tokens", "type": "uint256"},
            ],
            "name": "transfer",
            "outputs": [{"internalType": "bool", "name": "success", "type": "bool"}],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "constant": False,
            "inputs": [
                {"internalType": "address", "name": "tokenAddress", "type": "address"},
                {"internalType": "uint256", "name": "tokens", "type": "uint256"},
            ],
            "name": "transferAnyERC20Token",
            "outputs": [{"internalType": "bool", "name": "success", "type": "bool"}],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "constant": False,
            "inputs": [
                {"internalType": "address", "name": "from", "type": "address"},
                {"internalType": "address", "name": "to", "type": "address"},
                {"internalType": "uint256", "name": "tokens", "type": "uint256"},
            ],
            "name": "transferFrom",
            "outputs": [{"internalType": "bool", "name": "success", "type": "bool"}],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "constant": False,
            "inputs": [
                {"internalType": "address", "name": "_newOwner", "type": "address"}
            ],
            "name": "transferOwnership",
            "outputs": [],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {
            "constant": False,
            "inputs": [
                {"internalType": "address", "name": "_newProxy", "type": "address"}
            ],
            "name": "transferProxy",
            "outputs": [],
            "payable": False,
            "stateMutability": "nonpayable",
            "type": "function",
        },
        {"payable": True, "stateMutability": "payable", "type": "fallback"},
        {
            "constant": True,
            "inputs": [],
            "name": "_getBaseLockTimestamp",
            "outputs": [{"internalType": "uint128", "name": "", "type": "uint128"}],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [
                {
                    "internalType": "uint256",
                    "name": "_requestListItem",
                    "type": "uint256",
                }
            ],
            "name": "_getDORequest",
            "outputs": [
                {"internalType": "address", "name": "downer", "type": "address"},
                {"internalType": "uint8", "name": "cpuRequest", "type": "uint8"},
                {"internalType": "uint8", "name": "memoryRequest", "type": "uint8"},
                {"internalType": "uint8", "name": "storageRequest", "type": "uint8"},
                {"internalType": "uint8", "name": "bandwidthRequest", "type": "uint8"},
                {"internalType": "uint16", "name": "duration", "type": "uint16"},
                {"internalType": "uint8", "name": "maxPrice", "type": "uint8"},
                {"internalType": "uint256", "name": "status", "type": "uint256"},
            ],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [
                {
                    "internalType": "uint256",
                    "name": "_requestListItem",
                    "type": "uint256",
                }
            ],
            "name": "_getDORequestMetadata",
            "outputs": [
                {"internalType": "address", "name": "downer", "type": "address"},
                {"internalType": "string", "name": "metadata1", "type": "string"},
                {"internalType": "string", "name": "metadata2", "type": "string"},
                {"internalType": "string", "name": "metadata3", "type": "string"},
                {"internalType": "string", "name": "metadata4", "type": "string"},
            ],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [],
            "name": "_getDORequestsCount",
            "outputs": [
                {"internalType": "uint256", "name": "_length", "type": "uint256"}
            ],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [
                {
                    "internalType": "uint256",
                    "name": "_requestListItem",
                    "type": "uint256",
                }
            ],
            "name": "_getDPRequest",
            "outputs": [
                {"internalType": "address", "name": "dproc", "type": "address"},
                {"internalType": "uint8", "name": "cpuRequest", "type": "uint8"},
                {"internalType": "uint8", "name": "memoryRequest", "type": "uint8"},
                {"internalType": "uint8", "name": "storageRequest", "type": "uint8"},
                {"internalType": "uint8", "name": "bandwidthRequest", "type": "uint8"},
                {"internalType": "uint16", "name": "duration", "type": "uint16"},
                {"internalType": "uint8", "name": "minPrice", "type": "uint8"},
                {"internalType": "uint256", "name": "status", "type": "uint256"},
            ],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [
                {
                    "internalType": "uint256",
                    "name": "_requestListItem",
                    "type": "uint256",
                }
            ],
            "name": "_getDPRequestMetadata",
            "outputs": [
                {"internalType": "address", "name": "dproc", "type": "address"},
                {"internalType": "string", "name": "metadata1", "type": "string"},
                {"internalType": "string", "name": "metadata2", "type": "string"},
                {"internalType": "string", "name": "metadata3", "type": "string"},
                {"internalType": "string", "name": "metadata4", "type": "string"},
            ],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [],
            "name": "_getDPRequestsCount",
            "outputs": [
                {"internalType": "uint256", "name": "_length", "type": "uint256"}
            ],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [
                {
                    "internalType": "uint256",
                    "name": "_requestListItem",
                    "type": "uint256",
                }
            ],
            "name": "_getDPRequestWithCreationDate",
            "outputs": [
                {"internalType": "address", "name": "dproc", "type": "address"},
                {"internalType": "uint8", "name": "cpuRequest", "type": "uint8"},
                {"internalType": "uint8", "name": "memoryRequest", "type": "uint8"},
                {"internalType": "uint8", "name": "storageRequest", "type": "uint8"},
                {"internalType": "uint8", "name": "bandwidthRequest", "type": "uint8"},
                {"internalType": "uint16", "name": "duration", "type": "uint16"},
                {"internalType": "uint8", "name": "minPrice", "type": "uint8"},
                {"internalType": "uint256", "name": "status", "type": "uint256"},
                {"internalType": "uint32", "name": "createdAt", "type": "uint32"},
            ],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [{"internalType": "address", "name": "addr", "type": "address"}],
            "name": "_getLockoutTimestamp",
            "outputs": [{"internalType": "uint128", "name": "", "type": "uint128"}],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [
                {
                    "internalType": "uint256",
                    "name": "_requestListItem",
                    "type": "uint256",
                }
            ],
            "name": "_getMetadataCountForDPRequest",
            "outputs": [
                {"internalType": "uint256", "name": "_length", "type": "uint256"}
            ],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [
                {
                    "internalType": "uint256",
                    "name": "_requestListItem",
                    "type": "uint256",
                }
            ],
            "name": "_getMetadataCountForRequest",
            "outputs": [
                {"internalType": "uint256", "name": "_length", "type": "uint256"}
            ],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [
                {
                    "internalType": "uint256",
                    "name": "_requestListItem",
                    "type": "uint256",
                },
                {"internalType": "uint256", "name": "_metadataItem", "type": "uint256"},
            ],
            "name": "_getMetadataValueForDPRequest",
            "outputs": [
                {"internalType": "string", "name": "key", "type": "string"},
                {"internalType": "string", "name": "value", "type": "string"},
            ],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [
                {
                    "internalType": "uint256",
                    "name": "_requestListItem",
                    "type": "uint256",
                },
                {"internalType": "uint256", "name": "_metadataItem", "type": "uint256"},
            ],
            "name": "_getMetadataValueForRequest",
            "outputs": [
                {"internalType": "string", "name": "key", "type": "string"},
                {"internalType": "string", "name": "value", "type": "string"},
            ],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [],
            "name": "_getMyDOOrders",
            "outputs": [
                {"internalType": "uint256[]", "name": "req", "type": "uint256[]"}
            ],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [],
            "name": "_getMyDORequests",
            "outputs": [
                {"internalType": "uint256[]", "name": "req", "type": "uint256[]"}
            ],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [],
            "name": "_getMyDPRequests",
            "outputs": [
                {"internalType": "uint256[]", "name": "req", "type": "uint256[]"}
            ],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [
                {"internalType": "uint256", "name": "_orderItem", "type": "uint256"}
            ],
            "name": "_getOrder",
            "outputs": [
                {"internalType": "address", "name": "downer", "type": "address"},
                {"internalType": "address", "name": "dproc", "type": "address"},
                {"internalType": "uint256", "name": "doRequest", "type": "uint256"},
                {"internalType": "uint256", "name": "dpRequest", "type": "uint256"},
                {"internalType": "uint256", "name": "status", "type": "uint256"},
            ],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [],
            "name": "_getOrdersCount",
            "outputs": [
                {"internalType": "uint256", "name": "_length", "type": "uint256"}
            ],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [
                {"internalType": "uint256", "name": "_orderItem", "type": "uint256"}
            ],
            "name": "_getResultFromOrder",
            "outputs": [
                {"internalType": "string", "name": "_Result", "type": "string"}
            ],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [],
            "name": "_totalSupply",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [
                {"internalType": "address", "name": "tokenOwner", "type": "address"},
                {"internalType": "address", "name": "spender", "type": "address"},
            ],
            "name": "allowance",
            "outputs": [
                {"internalType": "uint256", "name": "remaining", "type": "uint256"}
            ],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [
                {"internalType": "address", "name": "tokenOwner", "type": "address"}
            ],
            "name": "balanceOf",
            "outputs": [
                {"internalType": "uint256", "name": "balance", "type": "uint256"}
            ],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [],
            "name": "callerAddress",
            "outputs": [{"internalType": "address", "name": "", "type": "address"}],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [],
            "name": "decimals",
            "outputs": [{"internalType": "uint8", "name": "", "type": "uint8"}],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [],
            "name": "implementation",
            "outputs": [{"internalType": "address", "name": "", "type": "address"}],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [],
            "name": "implementationPro",
            "outputs": [{"internalType": "address", "name": "", "type": "address"}],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [],
            "name": "name",
            "outputs": [{"internalType": "string", "name": "", "type": "string"}],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [],
            "name": "newOwner",
            "outputs": [{"internalType": "address", "name": "", "type": "address"}],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [],
            "name": "owner",
            "outputs": [{"internalType": "address", "name": "", "type": "address"}],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [],
            "name": "symbol",
            "outputs": [{"internalType": "string", "name": "", "type": "string"}],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
        {
            "constant": True,
            "inputs": [],
            "name": "totalSupply",
            "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
            "payable": False,
            "stateMutability": "view",
            "type": "function",
        },
    ],
}



from setuptools import setup, find_packages

# read the contents of your README file
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (
    this_directory / "ethernity_cloud_runner_py" / "README.md"
).read_text()

print(this_directory);


setup(
    name="ethernity-cloud-runner-py",
    version="0.0.38",
    url="https://github.com/ethernity-cloud/ethernity-cloud-runner-py",
    author="Ethernity Cloud Team",
    author_email="contact@ethernity.cloud",
    description="Ethernity Cloud Runner Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages("."),
    install_requires=[
        "web3==6.19.0",
        "eth-typing==4.3.0",
        "ipfshttpclient",
        "python-dotenv",
        "tinyec",
        "cryptography",
        "requests",
        "pyasn1",
        "pynacl",
        "coincurve",
    ],
)
